/* Copyright (C) 2008 Laurent Perez <laurent.perez@unicaen.fr> */
/* Released under the terms of the GNU GPL (version 2 or later). */

#include <slang.h>
#include <sys/inotify.h>
#include <sys/ioctl.h>
#include <linux/types.h>

SLANG_MODULE (inotify);

static char *Module_Version_String = "0.1";
#define MODULE_VERSION_NUMBER (0*10000 + 1*100 + 0)

static int Inotify_Error = -1;

static void *sl_inotify_init (void)
{
   int fd;
   SLFile_FD_Type *f;
   
   fd = inotify_init ();
   if (fd < 0)
     {	
	SLang_verror (Inotify_Error, "Inotify init error.");
	return;
     }
   
   if (NULL == (f = SLfile_create_fd ("inotify", fd)))
     {
	SLang_verror (Inotify_Error, "Can't create file descriptor.");
	return;
     }   

   (void) SLfile_push_fd (f);
}

static int sl_inotify_add_watch (SLFile_FD_Type *f, const char *path, __u32 *mask)
{
   int fd;
   
   if (-1 == SLfile_get_fd (f, &fd))
     return -1;
   
   return inotify_add_watch (fd, path, *mask);
}

static int sl_inotify_rm_watch (SLFile_FD_Type *f, int *wd)
{
   int fd;

   if (-1 == SLfile_get_fd (f, &fd))
     return -1;
   
   return inotify_rm_watch (fd, *wd);
}

static int sl_inotify_pending (SLFile_FD_Type *f)
{
   unsigned int queue_len;
   int fd, retval;
   
   if (-1 == SLfile_get_fd (f, &fd))
     return -1;
   
   retval = ioctl (fd, FIONREAD, &queue_len);
   if (retval < 0)
     return -1;
   return queue_len;
}

static void sl_inotify_read_event (SLFile_FD_Type *f)
{
   char buf [sizeof (struct inotify_event) + 16];
   int fd, retval;
   struct inotify_event *event;
   
   SLfile_get_fd (f, &fd);
   
   retval = read (fd, buf, sizeof (struct inotify_event) + 16);
   if (retval < 0)
     {
	SLang_verror (Inotify_Error, "Read error.");
	return;
     }   
     
   event = (struct inotify_event *) &buf;
   
   SLang_push_int (event->wd);
   SLang_push_uint (event->mask);
   SLang_push_uint (event->cookie);
   if (event->len)
     SLang_push_string (event->name);
   else
     SLang_push_string ("");
}

#define I SLANG_INT_TYPE
#define U SLANG_UINT_TYPE
#define V SLANG_VOID_TYPE
#define F SLANG_FILE_FD_TYPE
#define S SLANG_STRING_TYPE
static SLang_Intrin_Fun_Type Module_Intrinsics [] =
{
   MAKE_INTRINSIC_0 ("inotify_init", sl_inotify_init, V),
   MAKE_INTRINSIC_3 ("inotify_add_watch", sl_inotify_add_watch, I, F, S, U),
   MAKE_INTRINSIC_2 ("inotify_rm_watch", sl_inotify_rm_watch, I, F, I),
   MAKE_INTRINSIC_1 ("inotify_pending", sl_inotify_pending, I, F),
   MAKE_INTRINSIC_1 ("inotify_read_event", sl_inotify_read_event, V, F),
   SLANG_END_INTRIN_FUN_TABLE
};
#undef I
#undef U
#undef V
#undef F
#undef S

static SLang_IConstant_Type Module_IConstants [] =
{
   MAKE_ICONSTANT ("IN_ACCESS", IN_ACCESS),
   MAKE_ICONSTANT ("IN_MODIFY", IN_MODIFY),
   MAKE_ICONSTANT ("IN_ATTRIB", IN_ATTRIB),
   MAKE_ICONSTANT ("IN_CLOSE_WRITE", IN_CLOSE_WRITE),
   MAKE_ICONSTANT ("IN_CLOSE_NOWRITE", IN_CLOSE_NOWRITE),
   MAKE_ICONSTANT ("IN_OPEN", IN_OPEN),
   MAKE_ICONSTANT ("IN_MOVED_FROM", IN_MOVED_FROM),
   MAKE_ICONSTANT ("IN_MOVED_TO", IN_MOVED_TO),
   MAKE_ICONSTANT ("IN_DELETE", IN_DELETE),
   MAKE_ICONSTANT ("IN_DELETE_SELF", IN_DELETE_SELF),
   MAKE_ICONSTANT ("IN_CLOSE", IN_CLOSE),
   MAKE_ICONSTANT ("IN_MOVE", IN_MOVE),
   MAKE_ICONSTANT ("IN_ALL_EVENTS", IN_ALL_EVENTS),
   MAKE_ICONSTANT ("IN_UNMOUNT", IN_UNMOUNT),
   MAKE_ICONSTANT ("IN_Q_OVERFLOW", IN_Q_OVERFLOW),
   MAKE_ICONSTANT ("IN_IGNORED", IN_IGNORED),
   MAKE_ICONSTANT ("IN_ISDIR", IN_ISDIR),
   MAKE_ICONSTANT ("IN_ONLYDIR", IN_ONLYDIR),
   MAKE_ICONSTANT ("IN_DONT_FOLLOW", IN_DONT_FOLLOW),
   MAKE_ICONSTANT ("IN_MASK_ADD", IN_MASK_ADD),
   MAKE_ICONSTANT ("IN_ISDIR", IN_ISDIR),
   MAKE_ICONSTANT ("IN_ONESHOT", IN_ONESHOT),
   SLANG_END_ICONST_TABLE
};

int init_inotify_module_ns (char *ns_name)
{
   SLang_NameSpace_Type *ns = SLns_create_namespace (ns_name);
   if (ns == NULL)
     return -1;
   
   if ((-1 == SLns_add_intrin_fun_table (ns, Module_Intrinsics, NULL))
       || (-1 == SLns_add_iconstant_table (ns, Module_IConstants, NULL)))
     return -1;

   if ((-1 == Inotify_Error)
       && (-1 == (Inotify_Error = SLerr_new_exception (SL_RunTime_Error,
						       "InotifyError", "Inotify Error"))))
     return -1;
   
   return 0;   
}
